#include <iostream>
#include <cstdlib>
#include <cctype>
#include <csignal>

#include "zero.hh"
#include "kertoimet.hh"
#include "vakiot.hh"
#include "yksikot.hh"
#include "yksikko.hh"
#include "summa.hh"
#include "tulo.hh"

extern void Dump(elemp res);

Kertoimet         Kert;
Yksikot           Yks;
Yksikkomuunnokset Ymu(Yks);
Vakiot            Vak(Yks);

class BadInput
{
	string desc;
public:
	BadInput() { }
	BadInput(const char *s) : desc(s) { }
	const string& value() const { return desc; }
};
class Parser
{
public:
    static void delblank(string &s)
    {
        while(s.size() && isspace(s[0]))s.erase(0, 1);
    }
    static merknum fetchnum(string &str) /* const is not valid for static */
    {
        char *ep;
        const char *s = str.c_str();
        const char *const begin = s;
        
        real val = strtold(s, &ep);
        
        // cout.precision(35);
        // cout << "value read: " << val << endl;
        
        /* Got the value, next analyze the precision */
        unsigned precis = 0;
        
        const bool VirallinenTyyli = true;
        
        const char *q = s;
        if(*q == '+' || *q == '-')++q;
        bool pilkku = false;
        unsigned nollia=0;
        for(; q < ep; ++q)
        {
            if(*q == '0') { if(pilkku&&(VirallinenTyyli||precis))++precis;else ++nollia; continue; }
            if(*q >= '1' && *q <= '9')
            {
                if(!pilkku)precis += nollia;
                ++precis;
                nollia = 0;
                continue;
            }
            /* Do not analyze the exponent */
            if(*q == 'E' || *q == 'e')break;
            if(*q == '.')
            {
                if(precis)precis += nollia;
                pilkku = true;
                nollia = 0;
            }
        }
        if(VirallinenTyyli || pilkku)
        {
            // Virallisessa 1000:n tarkkuus on 4
            // Joka tapauksessa 0.100:n tarkkuus on 3
            precis += nollia;
        }
        
        str.erase(0, ep-begin);
        delblank(str);
        
        return merknum(val, precis);
    }
    static tulo *ParseTulo(string &s, merknum &kerroin)
    {
        tulo *tulotmp = new tulo;
        
        bool jako = false;
        
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            unsigned a, b;
            int k = 0;
            
            elemp res;
            
            merknum ker = fetchnum(s);
            if(!ker.valid())ker = 1;
            
            a = Kert.Test(s);
            if(a && (b = Yks.Test(s, a)) && Yks.Value(s, a) != 'i') // Ei SI-kertoimia ille
            {
                k = Kert.Value(s);
                /* k:n "yksikk": 10^ */
            YksLabel:
                char y = Yks.Value(s, a);
                s.erase(0, a+b);
                
                ker.pow10(k);
                res = new yksikko(y);
            }
            else if((a = Vak.Test(s)))
            {
                res = &*Vak.Value(s);
                s.erase(0, a);
                
                ker.maxprecision();
            }
            else /* a=0 */
            {
                b = Yks.Test(s);
                if(b)
                {
                    /* k = 0; */
                    goto YksLabel;
                }
                /* a=0, b=0 */
                res = NULL;
                
                if(s.size() && s[0] == '(')
                {
                    s.erase(0, 1);
                    res = ParseSumma(s);
                    delblank(s);
                    if(s[0] != ')')throw BadInput("Expected: ')'");
                    s.erase(0, 1);
                }
            }
            
            real pot = 1;
            delblank(s);
            if(s.size())
            {
            	if(s[0]=='^')
	            {
    	            s.erase(0, 1);
        	        delblank(s);
	                pot = fetchnum(s).value();
    	        }
    	    }
    	    while(s.size())
    	    {
				if(s[0]!='' && s[0]!='')break;
				if(pot==1)pot=0;
				pot=pot*10;
				pot += (s[0]=='') ? 2 : 3;
  	        	s.erase(0, 1);
	    	}
            
            if(jako)
                kerroin /= powl(ker, pot);
            else
                kerroin *= powl(ker, pot);
                
            if(res)
                tulotmp->insert(res, jako ? -pot : pot);
            
            jako = false;
            if(s.size() && s[0]=='*')
            {
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='/')
            {
                jako = true;
                s.erase(0, 1);
            }
            else if(!res)break;
        }
        return tulotmp;
    }
    static elemp ParseSumma(string &s)
    {
        summa *summatmp = new summa;
        bool miinus = false;
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            merknum kerroin = 1;
            tulo *p = ParseTulo(s, kerroin);
            if(!p)break;
            
            if(p->size())
                summatmp->insert(p, miinus ? -kerroin : kerroin);
            else if(miinus)
                summatmp->lisays -= kerroin;
            else
                summatmp->lisays += kerroin;
            
            delblank(s);
            if(s.size() && s[0]=='+')
            {
                miinus = false;
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='-')
            {
                miinus = true;
                s.erase(0, 1);
            }
            else break;
        }
        return summatmp;
    }
    
    static void DumpMerkNum(const merknum &e)
    {
        char Buf[256];
        sprintf(Buf, "%.*Lf", (int)e.precision(), e.value());
        cout << Buf;
    }
    
    static void DumpResult(elemp result)
    {
        if(result->type=='+' || result->type=='*')
            cout << "\33[1;32m" << result->type << "\33[m";
        cout << flush;
        switch(result->type)
        {
            case '+':
            {
                summa &s = *(summa *)&*result;
                
                bool iniplus = false;
                
                if(s.lisays != 0 || s.elems.size() == 0)
                {
                    DumpMerkNum(s.lisays);
                    iniplus = true;
                }
                
                for(summa::elemap::const_iterator i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                    if(i->second >= 0)
                    {
                        if(iniplus)cout << " + ";
                        DumpMerkNum(i->second);
                    }
                    else
                    {
                        cout << " - ";
                        DumpMerkNum(-i->second);
                    }
                    DumpResult(i->first);
                    iniplus = true;
                }
                break;
            }
            case '*':
            {
                tulo &s = *(tulo *)&*result;
                
                tulo::elemap::const_iterator i;
                
                bool eka=true;
                
                for(i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                    if(i->second < 0)continue;
                    bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                    if(plus)cout << '(';
                    DumpResult(i->first);
                    if(plus)cout << ')';
                    if(i->second == 2)cout << '';
                    else if(i->second == 3)cout << '';
                    else if(i->second != 1)cout << '^' << i->second;
                    eka = false;
                }
                for(i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                    if(i->second >= 0)continue;
                    if(eka) { cout << '1'; eka = false; }
                    cout << '/';
                    bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                    if(plus)cout << '(';
                    DumpResult(i->first);
                    if(plus)cout << ')';
                    if(i->second == -2)cout << '';
                    else if(i->second == -3)cout << '';
                    else if(i->second != -1)cout << '^' << i->second;
                }
                
                if(!s.size())cout << '1';
                
                break;
            }
            
            default:
                cout << Yks[result->type];
        }
        cout << flush;
    }
    
    static void Parse(string s)
    {
        elemp result;
        bool vihre = false;
        try
        {
        	result = ParseSumma(s);
        }
        catch(BadInput &t)
        {
        	cout << "ce: " << t.value() << endl;
        	vihre = true;
        	result = NULL;
        }
        if(s.size())
        {
            cout << "ce: Parse error before \"" << s << "\"\n";
            vihre = true;
        }
        if(vihre)return;
        
        cout << "Before optimization: ";
        Dump(result);
        
        result = result->Expand();
        
        cout << "Result: ";        
        Dump(result);        
    }
};

static Parser p;
void Dump(elemp res)
{
    p.DumpResult(res);
    cout << endl;
}

static void SegTrap(int sig)
{
	cout << "Error: Unexpected input\n";
	raise(sig);
}
int main(int argc, const char *const *argv)
{
    string s;
    signal(SIGSEGV, SegTrap);
    for(int i=1; i<argc; i++)
    {
        if(strstr(argv[i], "help"))
        {
            cout << "Usage: " << argv[0] << "<expression>\n";
            return 0;
        }
        s += ' ';
        s += argv[i];
    }
    if(!s.size())
    {
        cout << "Anna lauseke: " << flush;
        getline(cin, s);
    }
    
    p.Parse(s);
    
    return 0;
}
